# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Revision $Id$

import os
import sys

import roslib.scriptutil

import inspect

from cStringIO import StringIO

import string

# Anything outside the scope of these primitives is a submessage
_PRIMITIVES = ['byte','int8','int16','int32','int64','char','uint8','uint16','uint32','uint64','float32','float64','string','time']

class BagMigrationException(Exception):
    pass



def clean_name(name, top_name):
    name_split = name.split('/')
    try:
        name_split.remove('roslib')
    except ValueError:
        pass
    try:
        name_split.remove(top_name.split('/')[0])
    except ValueError:
        pass

    new_name = string.join(name_split,'/')
    return new_name

class MessageRule(object):
    old_type = ""
    old_md5sum = ""
    old_full_text = ""
    new_type = ""
    new_md5sum = ""
    new_full_text = ""
    migrated_types = []

    valid = False
    
    def __init__(self, migrator):
        # Every rule needs to hang onto the migrator so we can potentially use it
        self.migrator = migrator

        # Instantiate types dynamically based on definition
        self.old_types = roslib.genpy.generate_dynamic(self.old_type, self.old_full_text)

        self.old_class = self.old_types[self.old_type]

        # Make sure our rule files are self consistent
        if (self.old_class._md5sum != self.old_md5sum):
            raise BagMigrationException("Bag migration rule for old_type: [%s] with md5sum [%s] has md5sum not matching definition"%(old_type, old_md5sum))


        if (self.new_full_text.strip() != ""):
            self.new_types = roslib.genpy.generate_dynamic(self.new_type, self.new_full_text)

            self.new_class = self.new_types[self.new_type]

            if (self.new_class._md5sum != self.new_md5sum):
                raise BagMigrationException("Bag migration rule for new_type [%s] with md5sum [%s] has md5sum not matching definition"%(new_type, new_md5sum))

        else:
            self.new_types = {}
            self.new_class = None

        # For any migrated type the user might want to use, we must
        # make sure the migrator had found a path for it It would be
        # nice not to need these through performing some kind of other
        # inspection of the update rule.
        for (t1,t2) in self.migrated_types:
            try:
                tmp_old_class = self.get_old_class(t1)
            except KeyError:
                raise BagMigrationException("Specified migrated type [%s] not found in old message types"%t1)
            try:
                tmp_new_class = self.get_new_class(t2)
            except KeyError:
                raise BagMigrationException("Specified migrated type [%s] not found in new message types"%t2)
            if (tmp_new_class._md5sum != tmp_old_class._md5sum):
                self.migrator.find_path(tmp_old_class, tmp_new_class)

    def get_new_class(self,t):
        try:
            try:
                return self.new_types[t]
            except KeyError:                
                return self.new_types['roslib/' + t]
        except KeyError:
            return self.new_types[self.new_type.split('/')[0] + '/' + t]

    def get_old_class(self,t):
        try:
            try:
                return self.old_types[t]
            except KeyError:                
                return self.old_types['roslib/' + t]
        except KeyError:
            return self.old_types[self.old_type.split('/')[0] + '/' + t]


    def migrate(self, msg_from, msg_to):
        tmp_msg_from = clean_name(msg_from._type, self.old_type)
        tmp_msg_to = clean_name(msg_to._type, self.new_type)
        if (tmp_msg_from, tmp_msg_to) not in self.migrated_types:
            raise BagMigrationException("Rule tried to perform a migration from old [%s] to new [%s] not listed in migrated_types"%(tmp_msg_from, tmp_msg_to))
        self.migrator.migrate(msg_from, msg_to)

    def migrate_array(self, msg_from_array, msg_to_array, msg_to_class):
        while len(msg_to_array) > 0:
            msg_to_array.pop()

        msg_to_array.extend( [msg_to_class() for i in xrange(len(msg_from_array))] )
        for i in xrange(len(msg_from_array)):
            self.migrate(msg_from_array[i], msg_to_array[i])

    def print_def(self):
        pass
        

# These are rules that involve fields changing to update the message
# will have to pass through a serialized form
class MessageUpdateRule(MessageRule):

    def apply(self, old_msg_in, raw_in=False, raw_out=False):
        if not self.valid:
            raise BagMigrationException("Attempted to apply an invalid rule")

        # If raw_in we must deserialize
        if raw_in:
            old_msg = self.old_class()
            old_msg.deserialize(old_msg)
        else:
            old_msg = old_msg_in

        # Apply update rule
        new_msg = self.new_class()
        self.update(old_msg, new_msg)

        # If raw_out we must serialize
        if raw_out:
            buff = StringIO()
            new_msg.serialize(buff)
            return (self.new_class, data)
        else:
            return new_msg
            
    def update(self, old_msg, new_msg):
        raise BagMigrationException("Tried to use rule without update overidden")

# These are rules applicable when the wire format hasn't actually
# changed.  These rules assume that even with different md5sums the
# two messages SERIALIZE to the same block of data
class MessageRenameRule(MessageRule):

    def apply(self, old_msg_in, raw_in=False, raw_out=False):
        if not self.valid:
            raise BagMigrationException("Attempted to apply an invalid rule")

        # If both raw (we are not chaining), life is easy:
        if raw_in and raw_out:
            return (self.new_class, old_msg_in[1])

        # If neither raw (implicit in a chain) it's reasonably fast to
        # use the serialization to do "copy"
        if (not raw_in) and (not raw_out):
            new_msg = self.new_class()
            buff = StringIO()
            old_msg_in.serialize(buff)
            new_msg.deserialize(buff.getvalue())
            return new_msg

        # If not raw_in we need to serialize into our outgoing buffer
        if (not raw_in) and raw_out:
            buff = StringIO()
            new_msg.serialize(buff)
            return (self.new_class, buff)
        
        # If not raw_out, we need to deserialize into our new msg
        if raw_in and (not raw_out):
            new_msg = self.new_class()
            new_msg.deserialize(old_msg_in[1].getvalue())
            return new_msg


class MessageMigrator(object):
    def __init__(self):
        self.rules = {}
        to_check = roslib.scriptutil.rospack_depends_on_1('rosbagmigration')
        for pkg in to_check:
            m_file = roslib.manifest.manifest_file(pkg, True)
            m = roslib.manifest.parse_file(m_file)
            p_rules = m.get_export('rosbagmigration', 'rule')
            for r in p_rules:
                try:
                    # load that packages namespace
                    roslib.load_manifest(pkg)
            
                    # import the specified rule module
                    mod = __import__(r)
                    for sub_mod in r.split('.')[1:]:
                        mod = getattr(mod, sub_mod)

                    for m in dir(mod):
                        m_attr = getattr(mod, m)
                        if inspect.isclass(m_attr) and (not m_attr == MessageRule) \
                                and (not m_attr == MessageRenameRule) \
                                and (not m_attr == MessageUpdateRule) \
                                and issubclass(m_attr, MessageRule):
                            ruleinst = m_attr(self)
#                            print "Loading rule to convert from: [%s](%s) to [%s](%s)"%(ruleinst.old_type,ruleinst.old_md5sum,ruleinst.new_type,ruleinst.new_md5sum)
                            self.add_rule(ruleinst)

                except ImportError:
                    print >> sys.stderr, "Cannot load rule file [%s] in package [%s]"%(r, pkg)
        
    def add_rule(self, r):
        k = (r.old_type, r.old_md5sum)
        if self.rules.has_key(k):
            raise BagMigrationException("Conflicting rules for updating [%s] [%s]"%k)
        else:
            self.rules[k] = r

    def all_rules_valid(self):
        return not False in [r.valid for (k,r) in self.rules.iteritems()]

    def print_invalid_rules(self):
        for (k,r) in self.rules.iteritems():
            if not r.valid:
                r.print_def()

    # This function determines the set of rules which must be created
    # to get from the old type to the new type.  These can then be
    # queried with new_rules().  Passing in None as the new type will
    # search through the current ROS tree for a match based on type
    # name.
    def find_path(self, old_class, new_class):
        # This most likely means an underspecified rename rule was passed in
        if old_class is None:
            return None

        # If new_class is None, see if there is a a message in the
        # system with the right name
        if new_class is None:
            new_class = roslib.scriptutil.get_message_class(old_class._type)

        # If we still couldn't find anything, we're going to need to
        # apply a rule.  A rename rule might get this to match our
        # system eventually, hence the recursion.
        if new_class is None:
            try:
                R = self.rules[(old_class._type, old_class._md5sum)]

                # Keep searching from this new point
                return self.find_path(R.new_class, None)

            # Otherwise make a rename rule, which means we don't have
            # something valid to do since we can't automagically
            # determine what message in our ROS install to rename to
            except KeyError:
                new_rule = self.make_rename_rule(old_class)
#                print "Making rule to convert from: [%s](%s)"%(new_rule.old_type,new_rule.old_md5sum)
                self.add_rule(new_rule(self))
                return None


        # If the md5sum matches the existing named definition, we're
        # fine and a migration call will succeed through
        # serialization/deserialization when migrate is called below.
        if (old_class._md5sum == new_class._md5sum):
            return new_class

        # If we have gotten to here, the md5sums don't match, so we
        # look for a rule or else make a new update rule.
        try:
            R = self.rules[(old_class._type, old_class._md5sum)]

            # Keep searching from this new point
            return self.find_path(R.new_class, new_class)
        except KeyError:
            new_rule = self.make_update_rule(old_class, new_class)
#            print "Making rule to convert from: [%s](%s) to [%s](%s)"%(new_rule.old_type,new_rule.old_md5sum,new_rule.new_type,new_rule.new_md5sum)
            R = new_rule(self)
            self.add_rule(R)
            return R.new_class

    def migrate(self, msg_from, msg_to):

        tmp_msg = msg_from

        # If messages haved matched md5sums, we handle it with serialization
        while (tmp_msg._md5sum != msg_to._md5sum):
            try:
                R = self.rules[(tmp_msg._type, tmp_msg._md5sum)]
            except KeyError:
                raise BagMigrationException("Could not find an update rule for [%s] [%s]"%(msg_from._type, msg_from._md5sum))
            if not R.valid:
                raise BagMigrationException("Found an invalid update rule for [%s] [%s]"%(msg_from._type, msg_from._md5sum))            
            tmp_msg = R.apply(tmp_msg)

        # Now that we have migrated, do the copy with a serialize/deserialize pair
        # There are likey more efficient ways to do this
        buff = StringIO()
        tmp_msg.serialize(buff)
        msg_to.deserialize(buff.getvalue())
            

    # This returns a template for a rename rule
    def make_rename_rule(self,old_class):
        name = "rename_%s"%old_class._md5sum

        # We assemble the class as a string and then exec it to end up with a class
        # that can essentially print its own definition.
        classdef = "class %s(MessageRenameRule):\n"%name
        classdef += "\told_type = \"%s\"\n"%old_class._type
        classdef += "\told_md5sum = \"%s\"\n"%old_class._md5sum
        classdef += "\told_full_text = \"\"\"%s\"\"\"\n"%old_class._full_text
        classdef += "\t#Populate the following fields with: rosrun rosbagmigration showfields.py MsgName\n"
        classdef += "\tnew_type = \"\"\n"
        classdef += "\tnew_md5sum = \"\"\n"
        classdef += "\tnew_full_text = \"\"\" \"\"\"\n"
        classdef += "\tvalid = False\n"
    
        printclassdef = classdef + "\tdef print_def(self):\n\t\tprint \'\'\'%s\'\'\'\n"%classdef

        # Is this a TERRIBLE idea?
        exec(printclassdef)
        return locals()[name]

    def make_update_rule(self,old_class, new_class):

        name = "update_%s"%old_class._md5sum

        # We assemble the class as a string and then exec it to end up with a class
        # that can essentially print its own definition.
        classdef = "class %s(MessageUpdateRule):\n"%name
        classdef += "\told_type = \"%s\"\n"%old_class._type
        classdef += "\told_md5sum = \"%s\"\n"%old_class._md5sum
        classdef += "\told_full_text = \"\"\"%s\"\"\"\n"%old_class._full_text
        classdef += "\tnew_type = \"%s\"\n"%new_class._type
        classdef += "\tnew_md5sum = \"%s\"\n"%new_class._md5sum
        classdef += "\tnew_full_text = \"\"\"%s\"\"\"\n"%new_class._full_text
        classdef += "\n"
    
        validdef = "\tvalid = True\n"

        updatedef = "\tdef update(self, old_msg, new_msg):\n"

        migratedefs = "\tmigrated_types = [\n"
        
        # Assign across primitives, self.migrate non-primitives
        for (s,t) in zip(new_class.__slots__, new_class._slot_types):
            warn_msg = None
            new_base_type, new_is_array, new_array_len = roslib.msgs.parse_type(t)
            try:
                ind = old_class.__slots__.index(s)
                old_base_type, old_is_array, old_array_len = roslib.msgs.parse_type(old_class._slot_types[ind])

                if (new_is_array != old_is_array):
                    warn_msg = "Could not match array with nonarray"

                elif (new_array_len != old_array_len):
                    warn_msg = "Fixed array lengths don't match"

                elif (new_base_type in _PRIMITIVES):
                    if (new_base_type != old_base_type):
                        warn_msg = "Primitive type changed"
                    else:
                        updatedef += "\t\tnew_msg.%s = old_msg.%s\n"%(s,s)

                else:
                    tmp_old_type = clean_name(old_base_type, old_class._type)
                    tmp_new_type = clean_name(new_base_type, new_class._type)

                    migratedefs += "\t\t(\"%s\",\"%s\"),"%(tmp_old_type, tmp_new_type)
                    if not new_is_array:
                        updatedef += "\t\tself.migrate(old_msg.%s, new_msg.%s)\n"%(s,s)
                    else:
                        updatedef += "\t\tself.migrate_array(old_msg.%s, new_msg.%s, self.get_new_class(%s))\n"%(s,s,t[:-2])
            except ValueError:
                warn_msg = "No matching field name in old message"

            if warn_msg is not None:
                validdef = "\tvalid = False\n"
                updatedef += "\t\t#%s\n"%warn_msg
                updatedef += "\t\tnew_msg.%s = %s\n"%(s,migration_default_value(t))

        migratedefs += "\n\t]\n"

        classdef += migratedefs + '\n' + validdef + '\n' + updatedef

        printclassdef = classdef +  "\tdef print_def(self):\n\t\tprint \'\'\'%s\'\'\'\n"%classdef
        
        # Is this a TERRIBLE idea?
        exec(printclassdef)
        return locals()[name]


def migration_default_value(field_type):
    if field_type in ['byte', 'int8', 'int16', 'int32', 'int64',\
                          'char', 'uint8', 'uint16', 'uint32', 'uint64']:
        return '0'
    elif field_type in ['float32', 'float64']:
        return '0.'
    elif field_type == 'string':
        # strings, byte[], and uint8s are all optimized to be strings
        return "''"
    elif field_type.endswith(']'): # array type
        base_type, is_array, array_len = roslib.msgs.parse_type(field_type)
        if base_type in ['byte', 'uint8']:
            # strings, byte[], and uint8s are all optimized to be strings
            if array_len is not None:
                return "chr(0)*%s"%array_len
            else:
                return "''"
        elif array_len is None: #var-length
            return '[]'
        else: # fixed-length, fill values
            def_val = migration_default_value(base_type)
            return '[' + ','.join(itertools.repeat(def_val, array_len)) + ']'
    else:
        return "self.new_types[%s]"%field_type

def load_rules():

    return rules


